"""
The Approximate Distance Map (ADM) index structure. Despite the name, this is
an exact indexing method, with the approximated distances being 
underestimated, and used for filtering using the triangle inequality.
"""
#print globals()
#__metaclass__ = type
import util
import pivotselector
from sets import Set as set
import random
#import MetricList


class ADM:
    
    def __init__(self,metric,numpivots=1,selector=pivotselector.RandomSelector):
        #First we find reference points
        data = metric
        dfunc = metric.getDistance
        pivots = selector(data,dfunc).generate(numpivots)
        n = len(data)
        self.adm = MetricList.generateList(n+1)
        self.min = MetricList.generateList(n+1)
        self.dfunc = dfunc
        self.data = data

        for ia,oa in enumerate(data):
            for ib,ob in enumerate(data):
                if ia == ib:
                    self.adm[ia][ib] = 0
                    self.min[ia][ib] = 0
                elif oa in pivots or ob in pivots:
                    if ia < ib:
                        dval = dfunc(oa,ob)
                        self.adm[ia][ib] = dval
                        self.min[ia][ib] = dval
                    else:
                        self.adm[ia][ib] = self.adm[ib][ia]
                        self.min[ia][ib] = self.min[ib][ia]
                else:
                    self.adm[ia][ib] = 0
                    self.min[ia][ib] = util.inf
        #Do a Ford-Fulkerson reduction:
        for k in xrange(n):
            for i in xrange(n):
                for j in xrange(n):
                    #print self.min[i][j]
                    self.adm[i][j] = max(self.adm[i][j],self.adm[i][k]-self.min[k][j],self.adm[j][k]-self.min[k][i])
                    self.min[i][j] = min(self.min[i][j],self.min[i][k]+self.min[k][j])
    
    def range_search(self,query,r):
       
        n = len(self.data)
        
        #Clean up adm and min (in case it has been searched before
        t = n+1
        
        for i in xrange(t):
            self.adm[n][i] = 0
            self.adm[i][n] = 0
            self.min[n][i] = util.inf
            self.min[i][n] = util.inf
        self.min[n][n] = 0
        candidates = range(n)
        result = []
        
        while candidates:
            #Select a candidate from candidates
            #Find the i so that ADM[t,i] is minimized
            
            randcands = []
            candmin = util.inf
            for i in candidates:
                if self.adm[n][i] < candmin:
                    randcands = [i]
                    candmin = self.adm[n][i]
                elif self.adm[n][i] == candmin:
                    randcands.append(i)
            
            cand = random.choice(randcands)
            
            dval = self.dfunc(self.data[cand],query)
            
            self.adm[n][cand] = dval
            self.adm[cand][n] = dval
            self.min[n][cand] = dval
            self.min[cand][n] = dval
           
            
            
            for i in xrange(n+1):
                #print i
                for j in xrange(n+1):
                    self.adm[i][j] = max(self.adm[i][j],
                                         dval-self.min[i][n]-self.min[cand][j],
                                         dval-self.min[i][cand]-self.min[j][n],
                                         self.adm[i][n]-dval-self.min[cand][j],
                                         self.adm[i][cand]-dval-self.min[n][j],
                                         self.adm[j][cand]-dval-self.min[i][n],
                                         self.adm[n][j]-dval-self.min[i][cand])
                    
                    self.min[i][j] = min(self.min[i][j],
                                         self.min[i][n]+dval+self.min[cand][j],
                                         self.min[i][cand]+dval+self.min[n][j])
            for object in candidates:
                #print "%d %f %f %f %f" %(object,self.min[object][n],self.adm[object][n],self.min[object][cand],self.adm[object][cand])
                if self.min[object][n] < r:
                    result.append(object)
                    candidates.remove(object)
                if self.adm[object][n] >= r:
                    candidates.remove(object)
            print len(candidates)
        return [self.data[x] for x in result]
    
USABLE = False
INDEX_NAME = "ADM"
INDEX_CLASS = ADM
OPTIMIZES_COMBINED_SEARCH = False
BIBTEX_KEYS = "Shasha:1989,Wang:1990"
