"""
The AESA index structure.
"""

from sets import Set as set
from util import argmin,DMArray

__metaclass__ = type
try:
    from javaimps import AESA
except:
    class AESA:
        
        def __init__(self, metric):
            self.data = list(metric)
            self.dfunc = metric.getDistance
            self.heur_data = {}
            self.set_up_matrix()
            self.set_up_heuristic()
    
        def set_up_matrix(self):
            U = self.data
            n = len(U)
            self.matrix = d = DMArray((n,n))
            for x in range(n):
                for y in range(n):
                    if x >= y: continue
                    # The arrays are quadratic anyway; might as well store both dirs:
                    d[x,y] = d[y,x] = self.dfunc(U[x], U[y])
    
        def set_up_heuristic(self):
            self.d_qp = {}
    
        def pivot_seen_hook(self, u, p, d):
            """
            Called to establish a relationship between the object u and a newly selected
            pivot p. The hook is designed to be used for updating information linked to u
            (such as the pivot permutations in iAESA). Should be overridden in subclasses.
            For data objects, the arguments are indices into the data set. The query is
            identified by the ID len(data).
            """
            # No need to update info about objects other than the query:
            if u < len(self.data): return
            self.d_qp[p] = d
    
        def heuristic(self, q, u, p, old):
            """
            Calculates the updated heuristic between q and u, with p being the new pivot
            and old being the previous heuristic value.
            """
            # The heuristic in AESA is simply a sum of lower bounds, so here we add
            # the lower bound |d(q,p)-d(u,p)| to the old sum. The distance d(q,p) is
            # found in self.heur_data[p] (rom pivot_seen_hook).
            d_qp = self.d_qp[p]
            d = self.matrix
            return abs(d_qp-d[u,p])
    
        def range_search(self, query, radius):
            # Note: In a neares-neighbor search, the maximum lower bound encountered for
            # each object must be stored, since it may exceed a future upper bound on the
            # search radius. For a fixed radius, each bound need only be checked once.
            #
            # The hook and heuristic code is used in order to make it simpler to implement
            # iAESA in a subclass.
            
            U = self.data
            n = len(U)
            d = self.matrix
    
            q = n                                # A unique ID for the query
            r = radius
            
            P = set() # Pivots
            C = set(range(n))
            r = radius
            result = set()
            D = {}                               # Heuristic (sum of lower bounds)
            
            hook = self.pivot_seen_hook
            h = self.heuristic
            
            for u in range(n):                   # No lower bounds seen yet
                D[u] = 0
            
            while C:                             # As long as there are candidates...
                p = argmin(D, C)                 # Heuristically find the best pivot, p
                d_qp = self.dfunc(query, U[p])   # Calculate d(q,p)
                P.add(p); C.remove(p)            # Move p to the pivot set
                hook(q, p, d_qp)
                if d_qp <= r:                     # Return it if it qualifies
                    result.add(p)
                F = set()
                for u in C:                      # For all candidates...
                    if abs(d_qp-d[u,p]) > r:     # Possibly filter out the candidate
                        F.add(u)
                    else:                        # Otherwise...
                        hook(u, p, d[u,p])
                        D[u] = h(q, u, p, D[u])  # Update the heuristic
                C -= F
                
            return [U[i] for i in result]

USABLE = True
INDEX_NAME = "AESA"
INDEX_CLASS = AESA
OPTIMIZES_COMBINED_SEARCH = False
BIBTEX_KEYS = "Vidal:1986,Vidal:1994,Rico-Juan:2002"
SQUARED_MEM = True
