"""
The CM-Tree index structure.
"""

__metaclass__ = type

from sets import Set as set
import util

class CMTree:

    
    def __init__(self, metric, maxNodeCapacity = 32, maxNodeSplitResult = 5, partitioningWeight = 0.8):
        self.dfunc = metric.getDistance
        self.root = Node(self.dfunc, maxNodeCapacity, maxNodeSplitResult, partitioningWeight)
        for o in metric:
            self.insert(o)
            #print self.range_search(8,1000)
        
    
    
    def range_search(self, query, range):
        return self.root.range_search(query, range)
    
    def insert(self, o):
        selectedLeaf = self.root.findAccommodatingLeaf(o)
        selectedLeaf.insert(Entry(o,0,None))
        if selectedLeaf.shouldSplitOverflow() or selectedLeaf != self.root and selectedLeaf.shouldSplitClustering(o):
            selectedLeaf.split()
            # The root object might have changed
            while self.root.parent is not None:
                self.root = self.root.parent
        elif selectedLeaf != self.root:
            selectedLeaf.updateNavigationObjects()
        
    
    
    
class Node:

    staticID = 0
    def __str__(self):
        return "Node ID "+str(self.id)+" Containing "+str(len(self.entries))+" entries ("+str(self.n)+")"    
    
    def __init__(self, dfunc, maxCapacity, T, W):
        self.id = Node.staticID
        Node.staticID = Node.staticID + 1
        self.type = 'leaf'
        self.dt = {}
        self.n = 0
        self.entries = []
        self.parent = None
        
        self.dfunc = dfunc
        self.maxCapacity = maxCapacity
        self.T = T
        self.W = W
    
    def findAccommodatingLeaf(self, o):
        if self.type == 'leaf':
            return self
        
        calcedDistances = [None]*self.n
        bestEntry = self.entries[0]
        calcedDistances[0] = self.dfunc(o, bestEntry.obj)
        bestDistance = calcedDistances[0]
        
        for i in xrange(1,self.n):
            entry = self.entries[i]
            
            canPrune = False
            for j in xrange(i):
                if calcedDistances[j] is None:
                    continue
                prevDist = calcedDistances[j]
                if bestDistance <= bestEntry.radius:
                    if abs(prevDist - self.d(j,i)) > entry.radius:
                        canPrune = True
                else:

                    if entry.radius < bestEntry.radius and abs(prevDist - self.d(j,i)) >= bestDistance:
                        canPrune = True
                    elif entry.radius <= bestEntry.radius and abs(prevDist - self.d(j,i)) > bestDistance:
                        canPrune = True

            if canPrune:
                continue
            
            calcedDistances[i] = self.dfunc(o, entry.obj)
            
            myRadiusExtension = max(0, calcedDistances[i] - entry.radius)
            bestRadiusExtension = max(0, bestDistance - bestEntry.radius)
            
            if (myRadiusExtension, calcedDistances[i], entry.radius) < (bestRadiusExtension, bestDistance, bestEntry.radius) :
                bestEntry = entry
                bestDistance = calcedDistances[i]
                
        
        return bestEntry.subtree.findAccommodatingLeaf(o)
                
                
    def insert(self, entry):
        for i in xrange(self.n):
            self.dt[(i,self.n)] = self.dfunc(self.entries[i].obj, entry.obj)
        self.entries.append(entry)
        self.n += 1
        
        
        
    def range_search(self, query, range, navigation_d = None):
        result = []
        dists = [None]*self.n
        
        # Figure out which index our navigation object is at
        if self.parent is None:
            navObj = None
        else:
            for entry in self.parent.entries:
                if entry.subtree is not self:
                    continue
                for i in xrange(self.n):
                    if self.entries[i].obj is entry.obj:
                        navObj = i
                    
        
        if navObj is not None:
            dists[navObj] = navigation_d
        
        for i in xrange(self.n):
            canBePruned = False
            for j in xrange(self.n):
                if dists[j] is None:
                    continue
                if abs(dists[j] - self.d(i,j)) > range + self.entries[i].radius:
                    canBePruned = True
            if canBePruned:
                continue
            if dists[i] is None:
                dists[i] = self.dfunc(self.entries[i].obj, query)
                if dists[i] > range + self.entries[i].radius:
                    continue
            if self.type == 'leaf':
                result += [self.entries[i].obj]
            else:
                result += self.entries[i].subtree.range_search(query, range, dists[i])
        return result            
        
    
    def d(self, a, b):
        if a == b:
            return 0
        if a > b:
            return self.d(b,a)
        return self.dt[(a,b)]
    
    def shouldSplitOverflow(self):
        if self.type == 'directory':
            return self.n > self.maxCapacity - self.T
        elif self.type == 'leaf':
            return self.n > self.maxCapacity - 1
        else:
            raise "Something has gone wrong, the node is neither a directory nor a leaf"
        
    def shouldSplitClustering(self, insertedObj):
        if self.n <= 5:
            # If there was fewer than 5 objects before, we have no basis to say
            # anything about clustering 
            return False
        
        navEntry = self.parent.getNavEntry(self)
        navIndex = -1
        for i in xrange(self.n):
            entry = self.entries[i]
            if entry.obj is insertedObj:
                insertedIndex = i
            if entry.obj is navEntry.obj:
                navIndex = i
        
        # Preventing error, not sure where it comes from yet
        if navIndex == -1:
            return False
        if navEntry.radius >= self.d(navIndex, insertedIndex):
            return False
        
        avg = 0
        sd = 0
        for i in xrange(self.n):
            if i != insertedIndex:
                avg += self.d(navIndex,i)
        
        avg /= self.n-2.0
        
        for i in xrange(self.n):
            if i != insertedIndex:
                sd += (self.d(navIndex,i)-avg)**2
        
        sd /= self.n-2.0
        sd = sd**0.5
        
        #If the value is more than 2 standard deviations away, a split is in order 
        return self.d(navIndex, insertedIndex) >= avg+2*sd
            
    
    def split(self):
        newRootAllocated = False
        if self.parent is None:
            newRoot = Node(self.dfunc, self.maxCapacity, self.T, self.W)
            newRoot.type = 'directory'
            self.parent = newRoot
            newRootAllocated = True
        
        partition = self.partition()
        
        nodes = []
        for group in partition:
            node = Node(self.dfunc, self.maxCapacity, self.T, self.W)
            node.parent = self.parent
            node.n = len(group)
            node.type = self.type
            for i in xrange(len(group)):
                node.entries.append(self.entries[group[i]])
                if self.type == 'directory':
                    node.entries[-1].subtree.parent = node
                for j in xrange(i+1, len(group)):
                    node.dt[(i,j)] = self.d(group[i], group[j])
                       
            nodes.append(node)
        
        if not newRootAllocated:
            self.parent.delete(self)
        
        postedObjects = []
        for node in nodes:
            (nav, pcr) = node.pcr()
            node.parent.insert(Entry(nav, pcr, node))
            postedObjects.append(nav)
        
        
        shouldSplitClustering = False
        if self.parent.parent is not None:
            for obj in postedObjects:
                if self.parent.shouldSplitClustering(obj):
                    shouldSplitClustering = True
                    break
        
        if shouldSplitClustering or self.parent.shouldSplitOverflow():
            self.parent.split()
        elif self.parent.parent is not None:
            self.parent.updateNavigationObjects()
            
        
    def delete(self, node):
        delIndex = None
        for i in xrange(self.n):
            if self.entries[i].subtree is node:
                delIndex = i
        
        if delIndex is None:
            raise "A subtree not present in node was attempted deleted"
        
        ndt = {}
        del self.entries[delIndex]
        self.n = self.n-1
        for i in xrange(self.n):
            for j in xrange(i+1, self.n):
                a = i
                if a >= delIndex:
                    a = a+1
                b = j
                if b >= delIndex:
                    b = b+1
                ndt[(i,j)] = self.d(a,b)
        self.dt = ndt        
        
    def updateNavigationObjects(self):
        if self.parent is None:
            return
        
        myNavEntry = self.parent.getNavEntry(self)
        
        (newNavObj, newPcr) = self.pcr()
        if newNavObj is not myNavEntry.obj:
            self.parent.delete(self)
            self.parent.insert(Entry(newNavObj, newPcr, self))
            self.parent.updateNavigationObjects()
        elif newPcr != myNavEntry.radius:
            prevCR = myNavEntry.radius
            myNavEntry.radius = newPcr
            if self.parent.parent is not None:
                parentNavEntry = self.parent.parent.getNavEntry(self.parent)
                if newPcr <= prevCR or parentNavEntry.radius < self.dfunc(parentNavEntry.obj, newNavObj) + newPcr:
                    self.parent.updateNavigationObjects()
            
    def getNavEntry(self, node):
        for entry in self.entries:
            if entry.subtree is node:
                return entry
        return None
        
    # Returns a partitioning of the entries, one element for each node, including this one    
    def partition(self):
        
        mdt = [[None]*self.n for _ in xrange(self.n)]
        taken = [False]*self.n
        groups = [[] for _ in xrange(self.n)]
        for i in xrange(self.n):
            groups[i].append(i)
            for j in xrange(i+1, self.n):
                mdt[i][j] = self.d(i,j)
                mdt[j][i] = mdt[i][j]
            
        
        candidatePartitions = []
        
        if self.n <= self.T:
            partition = self.getPartition(groups, taken)
            candidatePartitions.append((partition, self.spcr(partition), self.n))
        
        # Gradually reduce the number of sets in the partition, until we are left with only one
        for k in xrange(self.n-1, 1, -1):
            
            # Figure out which clusters to join
            bestDistance = util.inf
            for i in xrange(self.n):
                if taken[i]:
                    continue
                for j in xrange(i+1,self.n):
                    if not taken[j] and mdt[i][j] < bestDistance:
                        bestDistance = mdt[i][j]
                        a = i
                        b = j
            # Add all elements in b to a, and mark b as taken
            for e in groups[b]:
                groups[a].append(e)
            taken[b] = True
            
            # Update all distances in mdt
            for i in xrange(self.n):
                if taken[i] or i == a:
                    continue
                mdt[a][i] = max(mdt[a][i], mdt[b][i])
                mdt[i][a] = mdt[a][i]
            
            if k <= self.T:
                partition = self.getPartition(groups, taken)
                if k == self.T:
                    candidatePartitions.append((partition, self.spcr(partition), k))
                else:
                    (_,pspcr,_) = candidatePartitions[-1]
                    spcr = pspcr - self.gpcr(groups[b]) - self.gpcr(set(groups[a])-set(groups[b])) + self.gpcr(groups[a])
                    candidatePartitions.append((partition, spcr, self.n))
            
        return self.bestPartition(candidatePartitions)    
        
        
    def spcr(self, partition):
        ret = 0
        for group in partition:
            ret += self.gpcr(group)    
        return ret
        
    def gpcr(self, group):
        return self.pcr(group)[1]
    
    
    def pcr(self, group = None):
        if group is None:
            group = range(self.n)
        
        ret = (-1, util.inf)
        for a in group:
            pcr = self.entries[a].radius
            for b in group:
                if a == b:
                    continue
                pcr = max(pcr, self.d(a,b) + self.entries[a].radius + self.entries[b].radius)
            if pcr < ret[1]:
                ret = (self.entries[a].obj,pcr)
        return ret        
    
    # Returns the best partitioning of objects, input is given as a list of (partition, spcr(partition), len(partition)) triple
    def bestPartition(self, candidatePartitions):
        if len(candidatePartitions) == 1:
            return candidatePartitions[0][0] 
        maxspcr = -1
        minspcr = util.inf
        for (_,spcr,_) in candidatePartitions:
            maxspcr = max(maxspcr,spcr)
            minspcr = min(minspcr,spcr)
        
        bestGrade = -1
        
        for i in xrange(len(candidatePartitions)):
            (partition, spcr, n) = candidatePartitions[i]
            if maxspcr == minspcr:
                grade = (1-self.W)*(self.T-n)/(self.T-2)
            else:
                grade = self.W*(maxspcr-spcr)/(maxspcr-minspcr) + (1-self.W)*(self.T-n)/(self.T-2)
            if grade > bestGrade:
                bestGrade = grade
                bestPartition = partition
                
        return bestPartition        

    def getPartition(self, groups, taken):
        ret = []
        for i in xrange(len(taken)):
            if not taken[i]:
                ret.append(list(groups[i]))
        return ret        


    


class Entry:

    def __init__(self, obj, radius, subtree):
        self.obj = obj
        self.radius = radius
        self.subtree = subtree
    
    


USABLE = True
INDEX_NAME = "CM-Tree"
INDEX_CLASS = CMTree
OPTIMIZES_COMBINED_SEARCH = False
BIBTEX_KEYS = "Aronovich:2007"
