package bonsai.indexes;

import static bonsai.indexes.MTreeNodeType.DIRECTORY;
import static bonsai.indexes.MTreeNodeType.LEAF;

import java.util.ArrayList;
import java.util.List;
public abstract class AbstractMTreeNode<T> {
	
	public AbstractMTreeNode<T> parent;
	
	public final MTreeNodeType type;
	protected int capacity;
	protected AbstractMTreeEntry<T> nodeEntry;
	protected DistanceFunction df;
	protected List<AbstractMTreeEntry<T>> entries = new ArrayList<AbstractMTreeEntry<T>>();
	
	public abstract void insert(T o);
	public abstract List<T> rangeSearch(T o, double r);
	
	
	protected abstract void insert(AbstractMTreeEntry<T> entry);
	protected abstract void delete(AbstractMTreeEntry<T> entry);
	protected abstract AbstractMTreeNode<T> findAccommodatingLeaf(T o);
	protected abstract boolean shouldSplit();
	protected abstract List<AbstractMTreeNode<T>> getPartitions();
	protected abstract AbstractMTreeNode<T> createNewNode(MTreeNodeType type);
	protected abstract void updateNavigationObjects(boolean forceParentUpdate);
	
	
	protected AbstractMTreeNode(DistanceFunction df, int capacity, MTreeNodeType type){
		this.df = df;
		this.capacity = capacity;
		this.type = type;
		this.nodeEntry = createEntry(null, 0, this);
	}
	
	protected AbstractMTreeNode(DistanceFunction df, int capacity){
	    this(df, capacity, LEAF);
	}
	
	
	protected double getRadius(){
		return nodeEntry.radius;
	}
	protected void setRadius(double radius){
		nodeEntry.radius = radius;
	}
	protected T getPivot(){
		return nodeEntry.obj;
	}
	protected void setPivot(T pivot){
		nodeEntry.obj = pivot;
	}
	
	

	/**
	 *  Splits the node into the partitions returned by getPartitions()
	 */
	public void split(){
		boolean newRootAllocated = parent == null;
		if(newRootAllocated){
			parent = createNewNode(DIRECTORY);
		}else{
			parent.delete(nodeEntry);
		}
		
		List<AbstractMTreeNode<T>> partitions = getPartitions();
		for(AbstractMTreeNode<T> node: partitions){
			parent.insert(node.nodeEntry);
			node.parent = parent;
		}
		if(parent.shouldSplit()){
			parent.split();
		}else{
			parent.updateNavigationObjects(false);
		}
	}

	protected AbstractMTreeEntry<T> createEntry(T pivot, double radius, AbstractMTreeNode<T> subtree){
		return new AbstractMTreeEntry<T>(pivot, radius, subtree);
	}
	
	public String toString(){
	    List<String> res = getStringRepresentation();
	    StringBuilder ret = new StringBuilder(getPivot().toString()+ "(radius = "+getRadius()+")\n");
	    for(String s: res) ret.append(s+'\n');
	    return ret.toString();
	}
	
	private List<String> getStringRepresentation(){
	    List<String> ret = new ArrayList<String>();
	    for(int i = 0; i < entries.size(); i++){
	        char down = i == entries.size() - 1 ? ' ' : '|';
	        List<String> tmp = new ArrayList<String>();
	        String s = entries.get(i).obj.toString() + "(radius = "+entries.get(i).radius+(this instanceof MTreeNode ? ", pivDist = "+((MTreeEntry<T>)entries.get(i)).pivotDistance : "");
	        if(this instanceof PMTreeNode){
	        	s += " pds: ";
	        	PMTreeNode<T> me = (PMTreeNode<T>)this;
	        	PMTreeEntry<T> entry = (PMTreeEntry<T>)entries.get(i);
	        	for(int j = 0; j < me.pivots.size(); j++){
	        		s += me.pivots.get(j)+"("+entry.pdmin[j]+"-"+entry.pdmax[j]+") ";
	        	}
	        }
	        if(this instanceof MStarTreeNode){
	        	MStarTreeEntry<T> entry = (MStarTreeEntry<T>)entries.get(i);
	        	s += ", NN = " + (entry.NN == null ? "-1" : entry.NN.obj);
	        	s += ", NNdist = " + entry.NNdist; 
	        }
	        tmp.add(s+")");
	        if(type == DIRECTORY) tmp.addAll(entries.get(i).subtree.getStringRepresentation());
	        ret.add("+-"+tmp.get(0));
	        for(int j = 1; j < tmp.size(); j++){
	            ret.add(down+" "+tmp.get(j));
	        }
	    }
	    return ret;
	}
	
}
