package bonsai.indexes;

import static bonsai.indexes.MTreeNodeType.LEAF;
import static bonsai.indexes.MTreeNodeType.DIRECTORY;

import java.util.ArrayList;
import java.util.List;



@SuppressWarnings("unchecked")
public class MTreeNode<T> extends AbstractMTreeNode<T> {

	
	public MTreeNode(DistanceFunction df, int capacity){
		super(df, capacity);
	}
	
	public MTreeNode(DistanceFunction df, int capacity, MTreeNodeType type){
	    super(df, capacity, type);
	}

	@Override
	protected void setPivot(T pivot) {
		((MTreeEntry<T>)nodeEntry).pivotDistance = -1;
		super.setPivot(pivot);
	}

	@Override
	public List<T> rangeSearch(T o, double r) {
		return rangeSearch(o, r, df.getDistance(o, getPivot()));
	}
	
	private List<T> rangeSearch(T o, double r, double pivotDistance){
		
		List<T> ret = new ArrayList<T>();
		for(int i = 0; i < entries.size(); i++){
			MTreeEntry<T> entry = (MTreeEntry<T>)entries.get(i);
			
			// Try to prune without computing the distance to the subtree's pivot
			if(Math.abs(pivotDistance - entry.pivotDistance) - entry.radius > r) continue;
			
			double d;
			if(entry.obj == getPivot()) d = pivotDistance;
			else d = df.getDistance(o, entry.obj);
			
			// Try to prune using the newly attained information about the distance to the query object 
			if(d - entry.radius > r) continue;
			
			// No pruning is possible for this entry
			if(type == LEAF) ret.add(entry.obj);
			else ret.addAll(((MTreeNode<T>)entry.subtree).rangeSearch(o, r, d));
		}
		return ret;
	}
	
	@Override
	protected AbstractMTreeNode<T> findAccommodatingLeaf(T o) {
		return findAccommodatingLeaf(o, -1);
	}
	
	private AbstractMTreeNode<T> findAccommodatingLeaf(T o, double pivotDistance){
		if(this.type == LEAF) return this;
		
		MTreeNode<T> best = null;
		double bestDistance = Double.POSITIVE_INFINITY;
		double bestRadiusEnlargement = Double.POSITIVE_INFINITY;

		for(AbstractMTreeEntry<T> entry: entries){
			double dist;
			if(pivotDistance != -1 && entry.obj == getPivot()) dist = pivotDistance;
			else dist = df.getDistance(entry.obj, o);
			
			double radiusEnlargement = Math.max(0, dist-entry.radius);
			if(radiusEnlargement < bestRadiusEnlargement || (radiusEnlargement == bestRadiusEnlargement && dist < bestDistance)){
				best = (MTreeNode<T>) entry.subtree;
				bestRadiusEnlargement = radiusEnlargement;
				bestDistance = dist;
			}
		}
		
		return best.findAccommodatingLeaf(o, bestDistance);
	}

	@Override
	public void insert(T o) {
		insert(createEntry(o, 0, null));
	}
	
	@Override
	public boolean shouldSplit() {
		return entries.size() >= capacity;
	}
	
	@Override
	public void updateNavigationObjects(boolean forceParentUpdate){
	    boolean newPivot = entries.size() != 0;
	    for(AbstractMTreeEntry<T> entry: entries){
	        if(entry.obj == getPivot()) newPivot = false;
	    }
	    
	    if(newPivot){
	        setPivot(entries.get(0).obj);
	        
	        ((MTreeEntry<T>)entries.get(0)).pivotDistance = 0;
	    }
	    
		double oldRadius = getRadius();
		setRadius(0);
		for(AbstractMTreeEntry<T> aentry: entries){
		    
		    MTreeEntry<T> entry = (MTreeEntry<T>)aentry;
		    
		    if((newPivot && aentry.obj != getPivot()) || entry.pivotDistance == -1) entry.pivotDistance = df.getDistance(getPivot(), entry.obj);
			
		    setRadius(Math.max(getRadius(), entry.pivotDistance + entry.radius));
		}
		if((forceParentUpdate || newPivot || oldRadius != getRadius()) && parent != null) parent.updateNavigationObjects(false);
	}

	@Override
	protected AbstractMTreeNode<T> createNewNode(MTreeNodeType type) {
		return new MTreeNode<T>(df, capacity, type);
	}

	@Override
	protected void delete(AbstractMTreeEntry<T> entry) {
		for(int i = 0; i < entries.size(); i++){
			if(entries.get(i) == entry){
				entries.remove(i);
				return;
			}
		}
	}

	/*
	 * Uses the mM_RAD_2 method of partitioning nodes.
	 */
	@Override
	protected List<AbstractMTreeNode<T>> getPartitions() {
		List<AbstractMTreeNode<T>> ret = null;
		
		double[][] dists = getAllToAllDistances();
		
		for(int p1 = 0; p1 < dists.length; p1++){
			for(int p2 = p1+1; p2 < dists.length; p2++){
				
				double r1 = entries.get(p1).radius;
				double r2 = entries.get(p2).radius;
				
				List<Integer> ind1 = new ArrayList<Integer>();
				ind1.add(p1);
				List<Integer> ind2 = new ArrayList<Integer>();
				ind2.add(p2);
				for(int i = 0; i < dists.length; i++){
					if(i == p1 || i == p2) continue;
					if(dists[p1][i] <= dists[p2][i]){
						ind1.add(i);
						r1 = Math.max(r1, dists[p1][i] + entries.get(i).radius);
					}else{
						ind2.add(i);
						r2 = Math.max(r2, dists[p2][i]) + entries.get(i).radius;
					}
				}
				r1 += entries.get(p1).radius;
				r2 += entries.get(p2).radius;
				if(ret == null || Math.max(r1, r2) < Math.max(ret.get(0).getRadius(), ret.get(1).getRadius())){
					ret = new ArrayList<AbstractMTreeNode<T>>();
					ret.add(createNode(p1, r1, ind1, dists));
					ret.add(createNode(p2, r2, ind2, dists));
				}
			}
		}
		return ret;
	}
	
	protected AbstractMTreeNode<T> createNode(int pivot, double radius, List<Integer> indexes, double[][] dists){
		MTreeNode<T> ret = (MTreeNode<T>)createNewNode(type);
		ret.setPivot(entries.get(pivot).obj);
		ret.setRadius(radius);
		for(int ind: indexes){
			MTreeEntry<T> entry = (MTreeEntry<T>)entries.get(ind);
			if(type == DIRECTORY) entry.subtree.parent = ret;
			entry.pivotDistance = dists[pivot][ind];
			ret.entries.add(entry);
		}
		ret.updateNavigationObjects(false);
		return ret;
	}
	
	
	protected double[][] getAllToAllDistances(){
		double[][] dists = new double[entries.size()][entries.size()];
		for(int i = 0; i < dists.length; i++){
			for(int j = i+1; j < dists.length; j++){

				MTreeEntry<T> a = (MTreeEntry<T>) entries.get(i);
				MTreeEntry<T> b = (MTreeEntry<T>) entries.get(j);
				
				double d;
				if(a.obj == getPivot()) d = b.pivotDistance;
				else if(b.obj == getPivot()) d = a.pivotDistance;
				else d = df.getDistance(a.obj, b.obj);
				
				dists[i][j] = d;
				dists[j][i] = d;
			}
		}
		return dists;
	}

	@Override
	protected void insert(AbstractMTreeEntry<T> entry) {
		((MTreeEntry<T>)entry).pivotDistance = -1;
		entries.add(entry);
		updateNavigationObjects(false);
	}

	@Override
	protected AbstractMTreeEntry<T> createEntry(T pivot, double radius, AbstractMTreeNode<T> node) {
		MTreeEntry<T> ret = new MTreeEntry<T>(pivot, radius, (MTreeNode<T>)node);
		return ret;
	}
	
	

}
