package bonsai.indexes;

import static bonsai.indexes.MTreeNodeType.DIRECTORY;
import static bonsai.indexes.MTreeNodeType.LEAF;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.PriorityQueue;

public class PMStarTreeNode<T> extends PMTreeNode<T> {
	
	
	public PMStarTreeNode(DistanceFunction df, int capacity, List<T> pivots){
		this(df, capacity, LEAF, pivots);
	}
	
	public PMStarTreeNode(DistanceFunction df, int capacity, MTreeNodeType type, List<T> pivots){
		super(df, capacity, type, pivots);
	}
	
	@SuppressWarnings("unchecked")
	@Override
	protected void setPivot(T pivot) {
		if(pivot == getPivot()) return;
		super.setPivot(pivot);

		List<PMStarTreeEntry<T>> expiredNNs = new ArrayList<PMStarTreeEntry<T>>();
		if(parent != null){
			PMStarTreeEntry<T> myEntry = (PMStarTreeEntry<T>)nodeEntry;
			
			myEntry.NNdist = Double.POSITIVE_INFINITY;
			myEntry.NN = null;
			
			for(int i = 0; i < parent.entries.size(); i++){
				PMStarTreeEntry<T> entry = (PMStarTreeEntry<T>)parent.entries.get(i);
				if(entry == myEntry) continue;
				
				double d = df.getDistance(entry.obj, myEntry.obj);
				
				if(d < myEntry.NNdist){
					myEntry.NNdist = d;
					myEntry.NN = entry;
				}
				
				if(d <= entry.NNdist){
					entry.NNdist = d;
					entry.NN = myEntry;
				}else if(entry.NN == myEntry) expiredNNs.add(entry);
			}
			((PMStarTreeNode<T>)parent).updateNNs(expiredNNs);
		}
	}
	
	protected void updateNNs(List<PMStarTreeEntry<T>> expiredNNs){
		for(PMStarTreeEntry<T> entry: expiredNNs){
			entry.NN = null;
			entry.NNdist = Double.POSITIVE_INFINITY;
		}
		for(int i = 0; i < expiredNNs.size(); i++){
			PMStarTreeEntry<T> one = expiredNNs.get(i);
			for(int j = 0; j < entries.size(); j++){
				if(expiredNNs.indexOf(entries.get(j)) >= i) continue;
				
				PMStarTreeEntry<T> two = (PMStarTreeEntry<T>)entries.get(j);
				double d;
				if(one.obj == getPivot() && two.pivotDistance != -1) d = two.pivotDistance;
				else if(two.obj == getPivot() && one.pivotDistance != -1) d = one.pivotDistance;
				else d = df.getDistance(one.obj, two.obj);
				
				if(d < one.NNdist){
					one.NNdist = d;
					one.NN = two;
				}
				if(d < two.NNdist){
					two.NNdist = d;
					two.NN = one;
				}
			}
		}
	}

	@Override
	protected void delete(AbstractMTreeEntry<T> entry) {
		super.delete(entry);
		List<PMStarTreeEntry<T>> expiredNNs = new ArrayList<PMStarTreeEntry<T>>();
		for(int i = 0; i < entries.size(); i++){
			PMStarTreeEntry<T> oentry = (PMStarTreeEntry<T>) entries.get(i);
			if(oentry.NN == entry){
				expiredNNs.add((PMStarTreeEntry<T>)entries.get(i));
			}
		}
		updateNNs(expiredNNs);
	}

	@SuppressWarnings("unchecked")
	@Override
	protected AbstractMTreeEntry<T> createEntry(T pivot, double radius,
			AbstractMTreeNode<T> node) {
		
		int np = pivots == null ? 0 : pivots.size();
		return new PMStarTreeEntry<T>(pivot, radius, (PMStarTreeNode<T>)node, np);
	}

	@Override
	protected AbstractMTreeNode<T> createNewNode(MTreeNodeType type) {
		return new PMStarTreeNode<T>(df, capacity, type, pivots);
	}
	
	
	protected int[][] getRNNGraph(){
		int[][] ret = new int[entries.size()][];
		int[] cnts = new int[entries.size()];
		HashMap<PMStarTreeEntry<T>, Integer> entryToInt = new HashMap<PMStarTreeEntry<T>, Integer>();
		int at = 0;
		for(int i = 0; i < entries.size(); i++){
			entryToInt.put((PMStarTreeEntry<T>)entries.get(i),at++);
		}
		for(int i = 0; i < entries.size(); i++){
			PMStarTreeEntry<T> entry = (PMStarTreeEntry<T>)entries.get(i);
			if(entry.NN != null) cnts[entryToInt.get(entry.NN)]++;
		}
		for(int i = 0; i < entries.size(); i++){
			ret[i] = new int[cnts[i]];
			cnts[i] = 0;
		}
		for(int i = 0; i < entries.size(); i++){
			PMStarTreeEntry<T> entry = (PMStarTreeEntry<T>)entries.get(i);
			if(entry.NN != null) ret[entryToInt.get(entry.NN)][cnts[entryToInt.get(entry.NN)]++] = i;
		}
		return ret;
	}
	
	protected LinkedList<Integer> getSacrificeQueue(final int[][] rnn){
		LinkedList<Integer> ret = new LinkedList<Integer>();
		PriorityQueue<Integer> pq = new PriorityQueue<Integer>(rnn.length, new Comparator<Integer>(){
			public int compare(Integer a, Integer b){
				return rnn[b].length-rnn[a].length;
			}
		});
		for(int i = 0; i < rnn.length; i++) pq.add(i);
		while(pq.size() > 0){
			ret.add(pq.poll());
		}
		
		return ret;
	}

	@SuppressWarnings("unchecked")
	@Override
	protected List<T> rangeSearch(T o, double r, double pivotDist, double[] pds) {
		
		List<T> ret = new ArrayList<T>();
		boolean[] excluded = new boolean[entries.size()];
		int[][] rnn = getRNNGraph();
		LinkedList<Integer> que = getSacrificeQueue(rnn);
		
		node_search:
		while(que.size() > 0){
			int node = que.removeFirst();
			if(excluded[node]) continue;
			excluded[node] = true;
			
			PMStarTreeEntry<T> entry = (PMStarTreeEntry<T>) entries.get(node);
			// Ball exclusion without computing distance
			if(Math.abs(pivotDist - entry.pivotDistance) - entry.radius > r) continue node_search;
			
			// Pivot ring exclusion
			for(int j = 0; j < pivots.size(); j++){
				if(pds[j] - r > entry.pdmax[j]) continue node_search;
				if(pds[j] + r < entry.pdmin[j]) continue node_search;
			}
			
			// Compute the distance
			double d;
			if(entry.obj == getPivot()) d = pivotDist;
			else d = df.getDistance(o, entry.obj);

			for(int i = 0; i < rnn[node].length; i++){
				PMStarTreeEntry<T> other = (PMStarTreeEntry<T>)entries.get(rnn[node][i]);
				if(Math.abs(d - other.NNdist) - other.radius > r) excluded[rnn[node][i]] = true;
				else que.addFirst(rnn[node][i]);
			}
			
			// Try to prune using the real distance
			if(d - entry.radius > r) continue node_search;
			
			// We have to search/include the nodes
			if(type == LEAF) ret.add(entry.obj);
			else ret.addAll(((PMStarTreeNode<T>)entry.subtree).rangeSearch(o, r, d, pds));
			
		}
		
		return ret;
	}

	@Override
	protected void insert(AbstractMTreeEntry<T> aentry) {
		PMStarTreeEntry<T> entry = (PMStarTreeEntry<T>) aentry;
		entry.NN = null;
		entry.NNdist = Double.POSITIVE_INFINITY;
		for(int i = 0; i < entries.size(); i++){
			PMStarTreeEntry<T> other = (PMStarTreeEntry<T>)entries.get(i);
			double d = df.getDistance(entry.obj, other.obj);
			if(d < entry.NNdist){
				entry.NN = other;
				entry.NNdist = d;
			}
			if(d < other.NNdist){
				other.NN = entry;
				other.NNdist = d;
			}
		}
		super.insert(entry);
	}

	@SuppressWarnings("unchecked")
	@Override
	protected AbstractMTreeNode<T> createNode(int pivot, double radius,
			List<Integer> indexes, double[][] dists) {
		
		PMStarTreeNode<T> ret = (PMStarTreeNode<T>)createNewNode(type);
		ret.setPivot(entries.get(pivot).obj);
		ret.setRadius(radius);
		for(int ind: indexes){
			PMStarTreeEntry<T> entry = (PMStarTreeEntry<T>)entries.get(ind);
			if(type == DIRECTORY) entry.subtree.parent = ret;
			entry.pivotDistance = dists[pivot][ind];
			entry.NN = null;
			entry.NNdist = Double.POSITIVE_INFINITY;
			for(int i = 0; i < indexes.size(); i++){
				int ind2 = indexes.get(i);
				if(ind == ind2) continue;
				if(dists[ind][ind2] < entry.NNdist){
					entry.NN = (PMStarTreeEntry<T>)entries.get(ind2);
					entry.NNdist = dists[ind][ind2];
				}
			}
			ret.entries.add(entry);
		}
		ret.updateNavigationObjects(false);
		return ret;
	}
}
