package bonsai.indexes;

import static bonsai.indexes.MTreeNodeType.LEAF;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class PMTreeNode<T> extends MTreeNode<T> {

	protected List<T> pivots;
	
	public PMTreeNode(DistanceFunction df, int capacity, List<T> pivots){
		this(df, capacity, LEAF, pivots);
	}
	
	public PMTreeNode(DistanceFunction df, int capacity, MTreeNodeType type, List<T> pivots){
		super(df, capacity, type);
		this.pivots = pivots;
		this.nodeEntry = createEntry(null, 0, this);
	}

	@SuppressWarnings("unchecked")
	@Override
	protected AbstractMTreeEntry<T> createEntry(T pivot, double radius,
			AbstractMTreeNode<T> node) {
		int np = pivots == null ? 0 : pivots.size();
		return new PMTreeEntry<T>(pivot, radius, (PMTreeNode<T>)node, np);
	}

	@Override
	protected AbstractMTreeNode<T> createNewNode(MTreeNodeType type) {
		return new PMTreeNode<T>(df, capacity, type, pivots);
	}

	@Override
	public List<T> rangeSearch(T o, double r) {
		double[] pds = new double[pivots.size()];
		List<T> ret = new ArrayList<T>();
		for(int i = 0; i < pds.length; i++){
			pds[i] = df.getDistance(o, pivots.get(i));
			if(pds[i] <= r) ret.add(pivots.get(i));
		}
		ret.addAll(rangeSearch(o, r,df.getDistance(o, getPivot()),pds));
		return ret;
	}
	
	@SuppressWarnings("unchecked")
	protected List<T> rangeSearch(T o, double r, double pivotDist, double[] pds){
		
		List<T> ret = new ArrayList<T>();
		
		node_search:
		for(int i = 0; i < entries.size(); i++){
			PMTreeEntry<T> entry = (PMTreeEntry<T>) entries.get(i);
			
			// Ball exclusion without computing distance
			if(Math.abs(pivotDist - entry.pivotDistance) - entry.radius > r) continue node_search;
			
			// Pivot ring exclusion
			for(int j = 0; j < pivots.size(); j++){
				if(pds[j] - r > entry.pdmax[j]) continue node_search;
				if(pds[j] + r < entry.pdmin[j]) continue node_search;
			}
			
			// Compute the distance
			double d;
			if(entry.obj == getPivot()) d = pivotDist;
			else d = df.getDistance(o, entry.obj);
			
			// Try to prune using the real distance
			if(d - entry.radius > r) continue node_search;
			
			// We have to search/include the nodes
			if(type == LEAF) ret.add(entry.obj);
			else ret.addAll(((PMTreeNode<T>)entry.subtree).rangeSearch(o, r, d, pds));
		}
		
		return ret;
	}

	@Override
	public void updateNavigationObjects(boolean forceParentUpdate) {
		
		PMTreeEntry<T> myEntry = (PMTreeEntry<T>)nodeEntry;
		double[] oldpdmin = myEntry.pdmin.clone();
		double[] oldpdmax = myEntry.pdmax.clone();
		
		Arrays.fill(myEntry.pdmin, Double.POSITIVE_INFINITY);
		Arrays.fill(myEntry.pdmax, Double.NEGATIVE_INFINITY);
		for(int i = 0; i < entries.size(); i++){
			PMTreeEntry<T> entry = (PMTreeEntry<T>)entries.get(i);
			for(int j = 0; j < pivots.size(); j++){
				if(myEntry.pdmin[j] > entry.pdmin[j]){
					myEntry.pdmin[j] = entry.pdmin[j];
				}
				if(myEntry.pdmax[j] < entry.pdmax[j]){
					myEntry.pdmax[j] = entry.pdmax[j];
				}
			}
		}
		super.updateNavigationObjects(!Arrays.equals(oldpdmin, myEntry.pdmin) || !Arrays.equals(oldpdmax, myEntry.pdmax) || forceParentUpdate);
	}

	@Override
	public void insert(T o) {
		PMTreeEntry<T> entry = (PMTreeEntry<T>)createEntry(o, 0, null);
		for(int i = 0; i < pivots.size(); i++){
			double d = df.getDistance(o, pivots.get(i));
			entry.pdmin[i] = entry.pdmax[i] = d;
		}
		insert(entry);
	}
}
