package bonsai.indexes;

import static bonsai.indexes.MTreeNodeType.LEAF;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;
import java.util.PriorityQueue;
import java.util.Queue;
import java.util.Random;

public class SlimTreeNode<T> extends MTreeNode<T> {

	
	public SlimTreeNode(DistanceFunction df, int capacity){
		super(df, capacity);
	}
	
	public SlimTreeNode(DistanceFunction df, int capacity, MTreeNodeType type){
	    super(df, capacity, type);
	}
	
	
	@Override
	protected AbstractMTreeNode<T> createNewNode(MTreeNodeType type) {
		return new SlimTreeNode<T>(df, capacity, type);
	}

	@Override
	protected AbstractMTreeNode<T> findAccommodatingLeaf(T o) {
		return findAccommodatingLeaf(o, -1);
	}
	
	@SuppressWarnings("unchecked")
	private AbstractMTreeNode<T> findAccommodatingLeaf(T o, double pivotDistance){
		if(type == LEAF) return this;
		
		SlimTreeNode<T> best = null;
		double bestDistance = Double.POSITIVE_INFINITY;
		int bestSize = Integer.MAX_VALUE;
		for(int i = 0; i < entries.size(); i++){
			MTreeEntry<T> entry = (MTreeEntry<T>)entries.get(i);
			
			double dist;
			if(pivotDistance != -1 && entry.obj == getPivot()) dist = pivotDistance;
			else dist = df.getDistance(o, entry.obj);
			
			int size = dist <= entry.radius ? entry.subtree.entries.size() : Integer.MAX_VALUE;
			
			if(size < bestSize || (size == bestSize && dist < bestDistance)){
				best = (SlimTreeNode<T>) entry.subtree;
				bestDistance = dist;
				bestSize = size;
			}
		}
		
		return best.findAccommodatingLeaf(o, bestDistance);
	}
	
	

	@Override
	public void updateNavigationObjects(boolean forceParentUpdate) {
		if(entries.size() == 0){
			parent.delete(nodeEntry);
		}else{
			super.updateNavigationObjects(forceParentUpdate);
		}
	}



	private static Random rnd = new Random(1337);
	
	/**
	 * 
	 * @param addFrom	The node that is on the other side of the edge that connects this node to the tree.
	 * @param addCosts	The cost of the edge used to connect this node to the tree.
	 * @return	An array with the distances between the entries of the node, if calculated. The distances are -1 otherwise.
	 */
	protected double[][] getMST(int[] addFrom, double[] addCosts){
		int n = entries.size();
		int pivot = -1;
		for(int i = 0; i < n; i++) if(entries.get(i).obj == getPivot()) pivot = i;

		double[][] dists = getAllToAllDistances();
		boolean[] included = new boolean[n];
		for(int i = 0; i < n; i++) addCosts[i] = dists[pivot][i];
		Arrays.fill(addFrom, pivot);
		included[pivot] = true;
		
		for(int i = 0; i < n-1; i++){
			int who = -1;
			double bestCost = Double.POSITIVE_INFINITY;
			for(int j = 0; j < n; j++){
				if(included[j]) continue;
				if(addCosts[j] < bestCost){
					bestCost = addCosts[j];
					who = j;
				}
			}
			included[who] = true;
			for(int j = 0; j < n; j++){
				if(included[j]) continue;
				if(dists[who][j] < addCosts[j]){
					addCosts[j] = dists[who][j];
					addFrom[j] = who;
				}
			}
		}
		return dists;
	}
	
	@Override
	protected List<AbstractMTreeNode<T>> getPartitions() {

		int[] addFrom = new int[entries.size()];
		double[] addCosts = new double[entries.size()];
		
		if(df instanceof CountingDistanceFunction){
			((CountingDistanceFunction)df).setIsCounting(true);
		}
		double[][] dists = getMST(addFrom, addCosts);
		if(df instanceof CountingDistanceFunction){
			((CountingDistanceFunction)df).setIsCounting(false);
		}
		
		int largest = -1;
		double largestCost = -1;
		int tieCount = 0;
		for(int i = 0; i < addCosts.length; i++){
			if(addCosts[i] > largestCost){
				largestCost = addCosts[i];
				largest = i;
				tieCount = 1;
			}else if(addCosts[i] == largestCost){
				tieCount++;
				if(rnd.nextInt(tieCount) == 0){
					largest = i;
				}
			}
		}
		int other = addFrom[largest];
		addFrom[largest] = largest;
		List<AbstractMTreeNode<T>> ret = new ArrayList<AbstractMTreeNode<T>>();
		ret.add(getPartition(getIndexes(addFrom, largest), dists));
		ret.add(getPartition(getIndexes(addFrom, other), dists));
		return ret;
	}
	
	List<Integer> getIndexes(int[] addFrom, int node){
		List<Integer> indexes = new ArrayList<Integer>();
		indexes.add(node);
		for(int i = 0; i < indexes.size(); i++){
			int at = indexes.get(i);
			if(addFrom[at] != at){
				indexes.add(addFrom[at]);
				addFrom[at] = at;
			}
			for(int j = 0; j < addFrom.length; j++){
				if(addFrom[j] == j) continue;
				if(addFrom[j] == at){
					addFrom[j] = j;
					indexes.add(j);
				}
			}
		}
		return indexes;
	}
	
	AbstractMTreeNode<T> getPartition(List<Integer> indexes, double[][] dists){
				
		int pivot = -1;
		double radius = Double.POSITIVE_INFINITY;
		for(int i = 0; i < indexes.size(); i++){
			double r = 0;
			for(int j = 0; j < indexes.size(); j++){
				if(j == i) continue;
				r = Math.max(r, dists[indexes.get(i)][indexes.get(j)]+entries.get(indexes.get(j)).radius);
			}
			r += entries.get(indexes.get(i)).radius;
			if(r < radius){
				radius = r;
				pivot = indexes.get(i);
			}
		}
		boolean finished = true;
		for(int i = 0; i < indexes.size(); i++){
			if(dists[pivot][indexes.get(i)] == -1){
				double d = df.getDistance(entries.get(indexes.get(i)).obj, entries.get(pivot).obj);
				dists[pivot][indexes.get(i)] = dists[indexes.get(i)][pivot] = d;
				if(d > radius) finished = false;
			}
		}
		
		if(!finished) return getPartition(indexes, dists);
		
		return createNode(pivot, radius, indexes, dists);
	}
	
	public static final int MAX_SLIMDOWN_ITERATIONS = 10;
	@SuppressWarnings("unchecked")
	public void slimDown(){
		if(type == LEAF) return;
		
		for(AbstractMTreeEntry<T> entry: entries){
			((SlimTreeNode<T>)entry.subtree).slimDown();
		}
		
		boolean change = true;
		for(int t = 0; t < MAX_SLIMDOWN_ITERATIONS && change; t++){
			change = false;
			for(int i = 0; i < entries.size(); i++){
				SlimTreeNode<T> node = (SlimTreeNode<T>)entries.get(i).subtree;
				Queue<MTreeEntry<T>> children = node.getChildren();
				
				boolean furthest = true;
				outer:
				while(children.size() > 0){
					MTreeEntry<T> child = children.poll();
					for(int j = 0; j < entries.size(); j++){
						SlimTreeNode<T> sibling = (SlimTreeNode<T>)entries.get(j).subtree;
						// Avoid needing to split nodes because of this
						if(i == j || sibling.entries.size() >= capacity-1) continue;
						
						double d = df.getDistance(child.obj, sibling.getPivot());
						if(d+child.radius <= sibling.getRadius() && (furthest || d < child.pivotDistance)){
							node.delete(child);
							sibling.insert(child);
							change = true;
							node.updateNavigationObjects(false);
							sibling.updateNavigationObjects(false);
							continue outer;
						}
					}
					furthest = false;
				}
			}
		}
	}
	
	protected Queue<MTreeEntry<T>> getChildren(){
		PriorityQueue<MTreeEntry<T>> dists = 
			new PriorityQueue<MTreeEntry<T>>(entries.size()+1, new Comparator<MTreeEntry<T>>(){
				public int compare(MTreeEntry<T> a, MTreeEntry<T> b){
					double r1 = a.radius+a.pivotDistance;
					double r2 = b.radius+b.pivotDistance;
					return r1 < r2 ? 1 : r1 > r2 ? -1 : 0;
				}
			});
		for(int i = 0; i < entries.size(); i++){
			dists.add((MTreeEntry<T>)entries.get(i));
		}
		return dists;
	}

}
