"""
The LC (List of Clusters) index structure. Uses a fixed bucket size (because
fixed radius LC is harder to tune), and the p5 strategy (choose cluster
centers that maximize the distance to previous centers). Does not currently 
utilize an ancillary metric structure to speed up building.

The default bucket size is set to 12, the smallest of the two sizes tested in
the original publication, and the one of the two seemingly closest to the
optimum shown in the graphs of that paper. An alternate value can be supplied
through the `bucket_size` parameter to the constructor.
"""
# XXX Chavez and Navarro (2000, 2005) give more instructions on choosing
# bucket size. They use sqrt(n) (in 2000).

from util import bunch, DistQueue
#import heapq
import time
from sets import Set as set
__metaclass__ = type

class LC:
        
    def __init__(self, metric, bucket_size=12):
        self.bucket_size = bucket_size
        self.data = list(metric)
        self.dfunc = metric.getDistance
        self.build()
    
    def build(self):
        # Avoids some duplicate calls to dfunc:
        memo = {}
        def d(x, y):
            xn = id(x)
            yn = id(y)
            try:
                return memo[xn, yn]
            except KeyError:
                memo[xn, yn] = self.dfunc(x, y)
            return memo[xn, yn]
        centers = []
        
        # The tail recursion here can be removed (i.e., turned into a loop) if 
        # recursion depth becomes a problem.
        # --
        # It became a problem (arnebef)
        #
        data = self.data
        self.root = None
        next = None

        mem = [0]*len(data)
        
        #Smarter reimplmentation
        indices = set(range(len(self.data)))
        memlengths = [0]*len(data)
        while indices:
            #print len(indices)
            #Couldn't find a set.peek() function
            cmin = indices.pop() 
            indices.add(cmin)
            dmin = 0
            tt = time.time()
            if centers:
                for i in indices:
                    if memlengths[i] > dmin:
                        cmin = i
                        dmin = memlengths[i]
            c = data[cmin]
            centers.append(c)
            #print cmin
            indices.remove(cmin)
            for x in indices:
                mem[x] = self.dfunc(c,data[x])
                memlengths[x]+=mem[x]
            #print mem
            #dec = heapq.nsmallest(
            #        self.bucket_size,
            #        indices,
            #        key=lambda x: mem[x] 
            #        )
            mycmp = lambda a,b:mem[a]<mem[b] and -1 or mem[a]>mem[b]
            slist = list(indices)
            slist.sort(mycmp)
            dec = slist[:self.bucket_size]
            for i in dec: indices.remove(i)
            inside = [data[x] for x in dec]
            #print dec
            #print indices
            rc = len(inside) > 0 and max([d(c,x) for x in inside]) or 0
            t = bunch(center = c,radius=rc,inside=inside,outside = None)
            if not self.root:
                self.root = t
                next = self.root
            else:
                next.outside = t
                next = t



    def range_search(self, query, radius):
        result = []
        d = self.dfunc
        q = query
        r = radius
        def search(node):
            if node is None: return
            c = node.center
            rc = node.radius
            dc = d(c, q)
            if dc <= r:
                result.append(c)
            if dc <= rc + r:
                result.extend([x for x in node.inside if d(q,x) <= r])
            if dc > rc - r:
                search(node.outside)
        search(self.root)
        return result

USABLE = True
INDEX_NAME = "LC"
INDEX_CLASS = LC
OPTIMIZES_COMBINED_SEARCH = False
BIBTEX_KEYS = "Chavez:2000,Chavez:2005"
SQUARED_MEM = True

"""

        while data:
            cmin = data[0]
            dmin = 0
            if centers:
                # The p5 strategy: Maximize distance sum to previous centers:
                for ccur in data:
                    dcur = 0
                    for prev in centers:
                        dcur += d(ccur, prev)
                    if dcur > dmin:
                        dmin = dcur
                        cmin = ccur
            c = cmin
            centers.append(c)
            # Sort according to distance to c
            dec = sorted((x for x in data if x is not c),
                key=lambda x: d(c, x))
            inside, rest = dec[:self.bucket_size], dec[self.bucket_size:]
            rc = max(d(c, x) for x in inside)
            t = bunch(center = c,radius=rc,inside=inside,outside = None)
            data = rest
            if not self.root:
                self.root = t
                next = self.root
            else:
                next.outside = t
                next = t
"""


