"""
The M-Tree index structure.
"""

import util


__metaclass__ = type


class MTree:
    
    def __init__(self,metric, maxNodeCapacity=32):
       # print data
        self.dfunc = metric.getDistance
        self.root = Node(self.dfunc, maxNodeCapacity)
        for obj in metric:
            self.insert(obj)
            
    def insert(self, obj):
        leaf = self.root.findBestLeaf(obj)
        leaf.insert(Entry(obj, 0, None))
        if leaf.overflow():
            leaf.split()
            while self.root.parent is not None:
                self.root = self.root.parent
    
    def range_search(self, query, radius):
        #print str(query)+" "+str(radius)
        return self.root.range_search(query, radius)


class Node:
    
    staticID = 0
    def __init__(self, dfunc, maxNodeCapacity):
        self.dfunc = dfunc
        self.maxNodeCapacity = maxNodeCapacity
        self.id = Node.staticID
        Node.staticID += 1
        self.n = 0
        self.entries = []
        self.pivot = None
        self.type = 'leaf'
        self.parent = None
        
    def __str__(self):
        ret = "Node "+str(self.id)+" containing nodes ["
        for entry in self.entries:
            ret += str(entry.subtree.id)+", "
        ret += "]"
        return ret
    
    def range_search(self, query, range, dp = None):
        if dp is None:
            dp = self.dfunc(query, self.entries[self.pivot].obj)
        
        result = []
        for entry in self.entries:
            if abs(dp-entry.dp) > entry.radius + range:
                continue
            if entry.dp == 0:
                d = dp
            else:
                d = self.dfunc(entry.obj, query)
            if d > entry.radius + range:
                continue
            if self.type == 'leaf':
                result += [entry.obj]
            else:
                result += entry.subtree.range_search(query, range, d)
        return result
    
    def overflow(self):
        return self.n >= self.maxNodeCapacity
    
    def split(self):
        
        newRootAdded = False
        if self.parent is None:
            newRootAdded = True
            root = Node(self.dfunc, self.maxNodeCapacity)
            root.type = 'directory'
            self.parent = root
        
        dists = {}
        
        def d(a,b):
            if a == b:
                return 0
            if b > a: 
                return d(b,a)
            if a == self.pivot:
                return self.entries[b].dp
            if b == self.pivot:
                return self.entries[a].dp
            if not dists.has_key((a,b)):
                dists[(a,b)] = self.dfunc(self.entries[a].obj, self.entries[b].obj)
            return dists[(a,b)]
        
        best = (util.inf, None, None, None, None)
        
        for p1 in xrange(self.n):
            for p2 in xrange(p1):
                r1 = 0
                r2 = 0
                a1 = []
                a2 = []
                for i in xrange(self.n):
                    a = d(i,p1)
                    b = d(i,p2)
                    if a <= b:
                        r1 = max(r1, a)
                        a1.append((a,i))
                    else:
                        r2 = max(r2,b)
                        a2.append((b,i))
                cand = (r1+r2,p1,p2,a1,a2)
                if cand < best:
                    best = cand
                        
        if not newRootAdded:
            self.parent.remove(self)
            
        for i in xrange(1, 3, 1):
            node = Node(self.dfunc, self.maxNodeCapacity)
            node.parent = self.parent
            node.type = self.type
            node.n = len(best[i+2])
            for j in xrange(node.n):
                entry = self.entries[best[i+2][j][1]]
                entry.dp = best[i+2][j][0]
                if entry.subtree is not None:
                    entry.subtree.parent = node
                node.entries.append(entry)
                if(best[i+2][j][1] == best[i]):
                    node.pivot = j
            pentry = self.entries[best[i]]
            node.parent.insert(Entry(pentry.obj, node.getRadius(), node))
            
        if self.parent.overflow():
            self.parent.split()
        
            
                
                
    def remove(self, subtree):
        for i in xrange(self.n):
            if self.entries[i].subtree is subtree:
                toRemove = i
        
        del self.entries[toRemove]
        self.n -= 1
        if self.pivot == toRemove:
            self.pivot = None
        elif self.pivot > toRemove:
            self.pivot -= 1
        
        if self.parent is not None:
            self.parent.updateRadius(self)
        
    def updateRadius(self, subtree):
        oldRadius = self.getRadius()
        for i in xrange(self.n):
            if self.entries[i].subtree is subtree:
                if subtree.pivot is not None:
                    self.entries[i].obj = subtree.entries[subtree.pivot].obj
                self.entries[i].radius = subtree.getRadius()
        
        newRadius = self.getRadius()
        if newRadius != oldRadius and self.parent is not None:
            self.parent.updateRadius(self) 
    
    def getRadius(self):
        radius = 0
        for entry in self.entries:
            radius = max(radius, entry.radius + entry.dp)
        return radius
    
    def insert(self, entry):
        self.n += 1
        if self.pivot is not None:
            entry.dp = self.dfunc(self.entries[self.pivot].obj, entry.obj)
        else:
            entry.dp = 0
            self.pivot = self.n-1
            for i in xrange(self.n-1):
                self.entries[i].dp = self.dfunc(self.entries[i].obj, entry.obj)
        self.entries.append(entry)
        if self.parent is not None:
            self.parent.updateRadius(self)
    
    def findBestLeaf(self, obj, piv_dist = None):
        if self.type == 'leaf':
            return self
        
        dists = [None]*self.n
        if piv_dist is not None:
            dists[self.pivot] = piv_dist
        
        for i in xrange(self.n):
            if dists[i] is None:
                dists[i] = self.dfunc(obj, self.entries[i].obj)
        
        best = (util.inf, None, None)
        
        for i in xrange(self.n):
            cand = (max(0,dists[i]-self.entries[i].radius), dists[i], i)
            if cand < best:
                best = cand
        
        return self.entries[cand[2]].subtree.findBestLeaf(obj, cand[1]) 


class Entry:
    def __init__(self, obj, radius, subtree):
        self.obj = obj
        self.radius = radius
        self.subtree = subtree
        self.dp = 0
        
        
USABLE = False # Is usable, but slower than Java version
INDEX_NAME = "M-Tree"
INDEX_CLASS = MTree
OPTIMIZES_COMBINED_SEARCH = False
BIBTEX_KEYS = "Zezula:1996,Ciaccia:1997,Ciaccia:1997a,Ciaccia:1998" # XXX
