# -*- coding: latin-1 -*-
"""
A simple in-memory implementation of the probabilistic proximity index
structure described by Chàvez et al. [LNCS 3789, pp. 405--414]. The
goal is to achieve a high degree of recall with as few calculations of
the dissimilarity function as possible. The principle of the structure
is as follows:

Select a set of pivots from the data set at random. For each object in
the data set, calculate the dissimilarity to each of these pivots, and
store a sorted permutation of the pivots. To perform a query,
calculate a similar permutation for the query, and compare this
permutation to those for the data objects, by calculating a simple
correlation (Spearman's Footrule) and discarding objects for which the
correlation is too low.

The index is designed (as published) for probabilistic
k-nearest-neighbor search, and a fixed number of objects (greater than
k) with maximal correlations to the query are retrieved as candidates.

Note that the distance function need not be metric for the method to work 
well.
"""
__metaclass__ = type

from random import *
from util import F

class PPIndex:
    
    def permutation(self, item):
        """
        Finds the pivot permutation for the given object, that is, the
        distance order of the pivots (from closest to farthest), with
        respect to the object. To simplify calculating the permutation
        correlation, the permutations are represented as a list of
        positions in the permutation for each element, rather than a
        permutation of the pivot indices. So, if perm[0] = 2, that 
        means that pivot 0 is the second closest, not that pivot 2 is
        the closest.
        """
        p = sorted(range(self.k), key=lambda p: self.dist(item, p))
        perm = [0] * len(p)
        for i, x in enumerate(p):
            perm[x] = i
        return perm
    
    def __init__(self, data, dist, pivot_count=10):
        self.dist = dist
        self.k = pivot_count
        self.pivots = sample(data, pivot_count)
        self.data = [(x, self.permutation(x)) for x in data]

    def search(self, query, range=None, count=None):
        if range is not None:
            raise NotImplementedError('Range search not implemented')
        if count is None:
            raise Error('A count must be specified')
        return self.knn_search(query, count, count*10)
    
    def knn_search(self, query, k, searched):
        qperm = self.permutation(query)
        corr_data = [(F(qperm, perm), x) for (x, perm) in self.data]
        candidates = [(self.dist(query, x), x) for (c, x) in
                      corr_data[:searched]]
        candidates.sort()
        return [x for (d, x) in candidates[:k]]

USABLE = False
# XXX Not an exact index structure
#INDEX_NAME = "PP-Index"
#INDEX_CLASS = PPIndex
OPTIMIZES_COMBINED_SEARCH = False
BIBTEX_KEYS = "Chavez:2005a"