#!/usr/bin/env python
import unittest
from glob import glob
import os, sys, re, random
from os import path
from glob import fnmatch
import bonsai 
idxmodules = bonsai.idxmodules
from bonsai import util
from metrics import Euclidian
# Determines whether the source is checked for "notes to self":
DISABLE_NOTES_CHECK = True

class IndexTestCase(unittest.TestCase):
    
    def __init__(self, module, testName):
        super(IndexTestCase, self).__init__(testName)
        self.module = module
        self.set_up_data()
    
    def set_up_data(self):
        self.d = d = 5
        self.n = n = 50
        self.q = q = 10

        euc = Euclidian(d,2)

        self.data = euc.generateData(n)
        self.queries = euc.generateRandomQueries(q)
        self.func = euc.generateDfunc()
    
    def shortDescription(self):
        doc = getattr(self,"_TestCase__testMethodDoc",None)
        if doc:
            return self.module.INDEX_NAME + ': ' + doc.split('\n')[0].strip()
        return None
    
    def setUp(self):
        self.instance = self.module.INDEX_CLASS(self.data, self.func)
    
    def test_creation(self):
        'Creating an index instance'
        # Implicit, through setUp()
    
    def test_knn_search(self):
        'Performing a KNN search'
        if not hasattr(self.instance, 'knn_search'):
            return
        d = self.func
        count = 4
        for q in self.queries:
            
            def f(o):
                return d(q, o)
            
            count += 1
            # Could use heapq.nsmallest, but that doesn't support a key 
            # function. (Could still use DSU, of course...):
            expected = sorted(self.data, key=f)[:count]
            result = sorted(self.instance.knn_search(q, count), key=f)
            self.failUnlessEqual(len(result), len(expected))
            for x, y in zip(expected, result):
                # Checking for equality is no good for arrays, so we use d, to 
                # make this array-agnostic:
                self.failUnlessEqual(d(x, y), 0)
    
    def test_range_search(self):
        'Performing a range search'
        d = self.func
        r = .7 # Arbitrary
        for q in self.queries:
            expected = [o for o in self.data if d(q, o) <= r]
            result = self.instance.range_search(q, r)
            self.failUnlessEqual(len(expected), len(result))
            
            def f(o):
                return d(q, o)
            
            expected.sort(key=f)
            result = sorted(result, key=f)
            for x, y in zip(expected, result):
                self.failUnlessEqual(d(x, y), 0)
    


class SourceCheckTestCase(unittest.TestCase):
        
    # Coded in ROT 13 to avoid spurious "self-reference"
    tags = 'KKK SVKZR UNPX GBQB GOQ'.encode('rot-13').split()
    start = path.normpath(path.join(path.dirname(__file__), '..'))
    suffixes = 'py tex sty'.split()
    tag_pat = re.compile('|'.join(tags))
        
    def test_notes_to_self(self):
        "Checking for unresolved 'notes to self'"
        # For use on commandline during development (don't set permanently):
        if DISABLE_NOTES_CHECK or str(os.getenv('DISABLE_NOTES_CHECK')).lower() in \
            ['true', 'yes', '1']:
            return
        matched = []
        for root, dirs, files in os.walk(self.start):
            for suffix in self.suffixes:
                for name in fnmatch.filter(files, '*.'+suffix):
                    fullname = os.path.join(root, name)
                    text = open(fullname).read()
                    if self.tag_pat.search(text):
                        matched.append(fullname)
        self.failUnlessEqual([], matched)
    


class UtilTestCase(unittest.TestCase):
    
    def test_tree_structure(self):
        
        # Create a tree and a single subtree:
        t = util.Tree(key='Key 1')
        self.failUnlessEqual(len(t), 0)
        self.assertRaises(IndexError, t.__getitem__, 0)
        s = util.SubTree(t, key='Key 2')
        
        self.failUnlessEqual(len(t), 1)
        
        self.failUnless(t[0] is s)
        
        # Exercise the attributes a bit:
        self.failIf(t.has('foo'))
        self.failUnless(t.get('foo') is None)
        self.failUnlessEqual(t.get('foo', 'bar'), 'bar')
        t.set('foo', 42)
        self.failUnless(t.has('foo'))
        self.failUnlessEqual(t.get('foo'), 42)
        self.failUnlessEqual(t.get('foo', 'bar'), 42)
        self.failUnlessEqual(sorted(t.keys()), ['foo', 'key'])
        self.failUnlessEqual(sorted(t.items()), [('foo', 42),
                                ('key', 'Key 1')])
        
        # Create ten separate trees:
        ts = [util.Tree() for i in range(10)]
        
        # Add the first five:
        for i in range(5):
            t.append(ts[i])
        self.failUnlessEqual(len(t), 6)
        
        # Remove the initial subtree:
        s = t[0]
        del t[0]
        
        # Replace the first five subtrees with the last five:
        for i in range(5):
            self.failUnless(t[i] is ts[i])
            t[i] = ts[i+5]
            self.failUnless(t[i] is ts[i+5])
        
        self.failUnlessEqual(t.children, ts[5:])
        
        # Delete subtrees 2 and 3 (7 an 8 from the list):
        self.failUnlessEqual(t[2:4], ts[7:9])
        del t[2:4]
        self.failUnlessEqual(len(t), 3)
        
        # Insert subtrees 0 through 4 from the list after subtree 2:
        t[2:2] = ts[:5]
        self.failUnlessEqual(len(t), 8)
        
        # Clear it, append a tree and test insert/remove:
        t.clear()
        self.failUnlessEqual(len(t), 0)
        t.append(ts[0])
        t.append(ts[1])
        t.insert(1, ts[2])
        self.failUnlessEqual(t.children, [ts[0], ts[2], ts[1]])
        t.remove(ts[2])
        self.failUnlessEqual(t.children, [ts[0], ts[1]])        
    
    
    def test_gist(self):
        
        class BLikeTreePolicy:
            
            def assign(self, obj, tree):
                # XXX Is it really necessary to add limits here, though?
                tree.set('obj', obj) # XXX Could just use the limits here :)
                tree.set('lo', obj)
                tree.set('hi', obj)
            
            def hits(self, query, tree):
                obj = tree.get('obj')
                if tree.has('obj') and query[0] <= obj <= query[1]:
                    return [obj]
                else:
                    return []
            
            def consistent(self, query, tree):
                if tree.get('hi') is None:
                    # Root:
                    return True
                return query[0] <= tree.get('hi') and query[1] >= tree.get('lo')
            
            def combine(self, tree):
                tree.set('lo', min([child.get('lo') for child in tree]))
                tree.set('hi', max([child.get('hi') for child in tree]))
            
            def situate(self, obj, tree):
                def cost(tree):
                    exp = max(0, tree.get('lo')-obj)
                    exp += max(0, obj-tree.get('hi'))
                    return exp
                child = util.argmin(cost, tree)
                child.set('lo', min(child.get('lo'), obj))
                child.set('hi', max(child.get('hi'), obj))
                return child
            
            def redistribute(self, tree, sibling):
                n = len(tree)
                sibling[:] = tree[n//2:]
                del tree[n//2:]
        
        # XXX How does/should object promotion work in general? Perhaps postprocessing
        # with the adjust method isn't the right way to go? A more proactive promotion
        # during the split might be necessary.
        # XXX How can/should the root be initialized? Should it simply be None to begin with?
        # Otherwise, it must be prepped by the policy.
        tree = util.GiST(BLikeTreePolicy(), capacity=3)
        # XXX Cheating:
        #tree.root.set('hi', 0) # XXX
        #tree.root.set('lo', 0) # XXX
        COUNT=100
        #COUNT=20
        for i in range(COUNT):
            tree.add(i)
        def dump(tree, ind=0):
            print (' '*ind*4) + "Tree", tree.get('lo'), tree.get('hi'), tree.get('type'), "{"
            for child in tree:
                dump(child, ind+1)
            print (' '*ind*4) + "}"
        
        #print; dump(tree.root)
        self.assertEqual(list(tree.search([25,30])), range(25,31))
        # XXX Add more; e.g., checking how the number of levels grows.
    


def gist_heights(gist):
    """
    Returns the minimum and maximum node depths of a tree.
    """
    heights = [None, None]
    def visit(tree, height=0):
        if tree.get('type') == 'object':
            if heights[0] is None or height < heights[0]:
                heights[0] = height
            if heights[1] is None or height < heights[1]:
                heights[1] = height
        for child in tree:
            visit(child, height+1)
    
    visit(gist.root)
    return heights


# XXX The standard IndexTestCase should *also* include circumstances (e.g., large enough data sets) where there is splitting involved.
class WhiteBoxTestCase(unittest.TestCase):
    """
    Test certain implementation/internal structure details of some of the index structures.
    """
    
    def test_mtree_balance(self):
        # XXX Should perhaps be a part of the index test case, and/or have access to the
        # standard data sets w/distances
        # XXX Should perhaps check more than just the balancing; could check some other 
        # whitebox-stuff
        data = range(1000)
        def dist(x, y):
            return abs(x-y)
        from bonsai import mtree
        if not mtree.USABLE: return
        gist = mtree.MTree(data, dist, 3).tree
        print gist_heights(gist) # XXX
        #self.assertEquals(util.tree_heights(tree), ) # TBD: What should it be?
        self.fail() # XXX

def suite(args=[]):
    # XXX Use args to filter tests/modules/structures
    loader = unittest.TestLoader()
    suite = unittest.TestSuite()
    mnames = [n for n in dir(IndexTestCase) if n.startswith('test')]
    for module in idxmodules:
        if not module.USABLE:
            continue
        for name in mnames:
            test = IndexTestCase(module, name)
            suite.addTest(test)
    suite.addTest(SourceCheckTestCase('test_notes_to_self'))
    suite.addTest(loader.loadTestsFromTestCase(WhiteBoxTestCase))
    suite.addTest(loader.loadTestsFromTestCase(UtilTestCase))
    return suite


class TestProgram(unittest.TestProgram):
        
    # Somewhat hackish... (copied from unittest.py)
    def parseArgs(self, argv):
        import getopt
        try:
            options, args = getopt.getopt(argv[1:], 'hHvq',
                                          ['help','verbose','quiet'])
            for opt, value in options:
                if opt in ('-h','-H','--help'):
                    self.usageExit()
                if opt in ('-q','--quiet'):
                    self.verbosity = 0
                if opt in ('-v','--verbose'):
                    self.verbosity = 2
            # Test loading/test name specification code deleted
            self.createTests(args)
        except getopt.error, msg:
            self.usageExit(msg)
    
    def createTests(self, args=[]):
        self.test = suite(args)
    


def test():
    TestProgram(module=None)

def main():
    test()

if __name__ == '__main__':
    main()
